/* 
 * This file is part of the iothpy library: python support for ioth.
 * 
 * Copyright (c) 2020-2024   Dario Mylonopoulos
 *                           Lorenzo Liso
 *                           Francesco Testa
 * Virtualsquare team.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#define PY_SSIZE_T_CLEAN
#include <Python.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <nlinline+.h>
#include <linux/if_link.h>

#define ADD_INT(name, value) \
    do { \
        if (PyModule_AddIntConstant(module, name, value) < 0) { \
            return NULL; \
        } \
    } while (0)

#define ADD_ENUM(names, n) \
    do { \
        for (int i = 0; i <= n; i++){ \
            ADD_INT(names[i],i); \
        } \
    } while(0)

PyDoc_STRVAR(const_linkadd_doc,
"This module extends the link_add settings in C to Python.\n\
Is needed to use the link_add method of stack.\n\
");



static const char* IFLA[] = {
	"IFLA_UNSPEC",
	"IFLA_ADDRESS",
	"IFLA_BROADCAST",
	"IFLA_IFNAME",
	"IFLA_MTU",
	"IFLA_LINK",
	"IFLA_QDISC",
	"IFLA_STATS",
	"IFLA_COST",
	"IFLA_PRIORITY",
	"IFLA_MASTER",
	"IFLA_WIRELESS",		/* Wireless Extension event - see wireless.h */
	"IFLA_PROTINFO",		/* Protocol specific information for a link */
	"IFLA_TXQLEN",
	"IFLA_MAP",
	"IFLA_WEIGHT",
	"IFLA_OPERSTATE",
	"IFLA_LINKMODE",
	"IFLA_LINKINFO",
	"IFLA_NET_NS_PID",
	"IFLA_IFALIAS",
	"IFLA_NUM_VF",		/* Number of VFs if device is SR-IOV PF */
	"IFLA_VFINFO_LIST",
	"IFLA_STATS64",
	"IFLA_VF_PORTS",
	"IFLA_PORT_SELF",
	"IFLA_AF_SPEC",
	"IFLA_GROUP",		/* Group the device belongs to */
	"IFLA_NET_NS_FD",
	"IFLA_EXT_MASK",		/* Extended info mask, VFs, etc */
	"IFLA_PROMISCUITY",	/* Promiscuity count: > 0 means acts PROMISC */
	"IFLA_NUM_TX_QUEUES",
	"IFLA_NUM_RX_QUEUES",
	"IFLA_CARRIER",
	"IFLA_PHYS_PORT_ID",
	"IFLA_CARRIER_CHANGES",
	"IFLA_PHYS_SWITCH_ID",
	"IFLA_LINK_NETNSID",
	"IFLA_PHYS_PORT_NAME",
	"IFLA_PROTO_DOWN",
	"IFLA_GSO_MAX_SEGS",
	"IFLA_GSO_MAX_SIZE",
	"IFLA_PAD",
	"IFLA_XDP",
	"IFLA_EVENT",
	"IFLA_NEW_NETNSID",
	"IFLA_IF_NETNSID",
	"IFLA_CARRIER_UP_COUNT",
	"IFLA_CARRIER_DOWN_COUNT",
	"IFLA_NEW_IFINDEX",
	"IFLA_MIN_MTU",
	"IFLA_MAX_MTU",
	"IFLA_PROP_LIST",
	"IFLA_ALT_IFNAME", /* Alternative ifname */
	"IFLA_PERM_ADDRESS",
	"IFLA_PROTO_DOWN_REASON",
	"IFLA_PARENT_DEV_NAME",
	"IFLA_PARENT_DEV_BUS_NAME",
	"IFLA_GRO_MAX_SIZE",
	"IFLA_TSO_MAX_SIZE",
	"IFLA_TSO_MAX_SEGS",
	"IFLA_ALLMULTI",		/* Allmulti count: > 0 means acts ALLMULTI */
	"IFLA_DEVLINK_PORT",
	"IFLA_GSO_IPV4_MAX_SIZE",
	"IFLA_GRO_IPV4_MAX_SIZE",
	"IFLA_DPLL_PIN",
	"__IFLA_MAX"
};
static const char* IFLA_PROTO_DOWNREASON[] = {
	"IFLA_PROTO_DOWN_REASON_UNSPEC",
	"IFLA_PROTO_DOWN_REASON_MASK",	/* u32, mask for reason bits */
	"IFLA_PROTO_DOWN_REASON_VALUE",   /* u32, reason bit value */
	"__IFLA_PROTO_DOWN_REASON_CNT",
};

static const char* IFLA_INET[] = {"IFLA_INET_UNSPEC", "IFLA_INET_CONF", "__IFLA_INET_MAX"};

/* Subtype attributes for IFLA_PROTINFO */
static const char* IFLA_INET6[] = {
	"IFLA_INET6_UNSPEC",
	"IFLA_INET6_FLAGS",	/* link flags			*/
	"IFLA_INET6_CONF",	/* sysctl parameters		*/
	"IFLA_INET6_STATS",	/* statistics			*/
	"IFLA_INET6_MCAST",	/* MC things. What of them?	*/
	"IFLA_INET6_CACHEINFO",	/* time values and max reasm size */
	"IFLA_INET6_ICMP6STATS",	/* statistics (icmpv6)		*/
	"IFLA_INET6_TOKEN",	/* device token			*/
	"IFLA_INET6_ADDR_GEN_MODE", /* implicit address generator mode */
	"IFLA_INET6_RA_MTU",	/* mtu carried in the RA message */
	"__IFLA_INET6_MAX"
};
static const char* IFLA_INET6_GEN_MODE[] = {"IN6_ADDR_GEN_MODE_EUI64",
	"IN6_ADDR_GEN_MODE_NONE", 
	"IN6_ADDR_GEN_MODE_STABLE_PRIVACY",
	"IN6_ADDR_GEN_MODE_RANDOM"
};

/* Bridge section */
static const char* IFLA_BR[] = { 
	"IFLA_BR_UNSPEC",
	"IFLA_BR_FORWARD_DELAY",
	"IFLA_BR_HELLO_TIME",
	"IFLA_BR_MAX_AGE",
	"IFLA_BR_AGEING_TIME",
	"IFLA_BR_STP_STATE",
	"IFLA_BR_PRIORITY",
	"IFLA_BR_VLAN_FILTERING",
	"IFLA_BR_VLAN_PROTOCOL",
	"IFLA_BR_GROUP_FWD_MASK",
	"IFLA_BR_ROOT_ID",
	"IFLA_BR_BRIDGE_ID",
	"IFLA_BR_ROOT_PORT",
	"IFLA_BR_ROOT_PATH_COST",
	"IFLA_BR_TOPOLOGY_CHANGE",
	"IFLA_BR_TOPOLOGY_CHANGE_DETECTED",
	"IFLA_BR_HELLO_TIMER",
	"IFLA_BR_TCN_TIMER",
	"IFLA_BR_TOPOLOGY_CHANGE_TIMER",
	"IFLA_BR_GC_TIMER",
	"IFLA_BR_GROUP_ADDR",
	"IFLA_BR_FDB_FLUSH",
	"IFLA_BR_MCAST_ROUTER",
	"IFLA_BR_MCAST_SNOOPING",
	"IFLA_BR_MCAST_QUERY_USE_IFADDR",
	"IFLA_BR_MCAST_QUERIER",
	"IFLA_BR_MCAST_HASH_ELASTICITY",
	"IFLA_BR_MCAST_HASH_MAX",
	"IFLA_BR_MCAST_LAST_MEMBER_CNT",
	"IFLA_BR_MCAST_STARTUP_QUERY_CNT",
	"IFLA_BR_MCAST_LAST_MEMBER_INTVL",
	"IFLA_BR_MCAST_MEMBERSHIP_INTVL",
	"IFLA_BR_MCAST_QUERIER_INTVL",
	"IFLA_BR_MCAST_QUERY_INTVL",
	"IFLA_BR_MCAST_QUERY_RESPONSE_INTVL",
	"IFLA_BR_MCAST_STARTUP_QUERY_INTVL",
	"IFLA_BR_NF_CALL_IPTABLES",
	"IFLA_BR_NF_CALL_IP6TABLES",
	"IFLA_BR_NF_CALL_ARPTABLES",
	"IFLA_BR_VLAN_DEFAULT_PVID",
	"IFLA_BR_PAD",
	"IFLA_BR_VLAN_STATS_ENABLED",
	"IFLA_BR_MCAST_STATS_ENABLED",
	"IFLA_BR_MCAST_IGMP_VERSION",
	"IFLA_BR_MCAST_MLD_VERSION",
	"IFLA_BR_VLAN_STATS_PER_PORT",
	"IFLA_BR_MULTI_BOOLOPT",
	"IFLA_BR_MCAST_QUERIER_STATE",
	"IFLA_BR_FDB_N_LEARNED",
	"IFLA_BR_FDB_MAX_LEARNED",
	"__IFLA_BR_MAX",
};

static const char* BRIDGE_MODE[]={"BRIDGE_MODE_UNSPEC","BRIDGE_MODE_HAIRPIN"};

static const char* IFLA_BRPORT[] = {
	"IFLA_BRPORT_UNSPEC",
	"IFLA_BRPORT_STATE",	/* Spanning tree state     */
	"IFLA_BRPORT_PRIORITY",	/* "             priority  */
	"IFLA_BRPORT_COST",	/* "             cost      */
	"IFLA_BRPORT_MODE",	/* mode (hairpin)          */
	"IFLA_BRPORT_GUARD",	/* bpdu guard              */
	"IFLA_BRPORT_PROTECT",	/* root port protection    */
	"IFLA_BRPORT_FAST_LEAVE",	/* multicast fast leave    */
	"IFLA_BRPORT_LEARNING",	/* mac learning */
	"IFLA_BRPORT_UNICAST_FLOOD", /* flood unicast traffic */
	"IFLA_BRPORT_PROXYARP",	/* proxy ARP */
	"IFLA_BRPORT_LEARNING_SYNC", /* mac learning sync from device */
	"IFLA_BRPORT_PROXYARP_WIFI", /* proxy ARP for Wi-Fi */
	"IFLA_BRPORT_ROOT_ID",	/* designated root */
	"IFLA_BRPORT_BRIDGE_ID",	/* designated bridge */
	"IFLA_BRPORT_DESIGNATED_PORT",
	"IFLA_BRPORT_DESIGNATED_COST",
	"IFLA_BRPORT_ID",
	"IFLA_BRPORT_NO",
	"IFLA_BRPORT_TOPOLOGY_CHANGE_ACK",
	"IFLA_BRPORT_CONFIG_PENDING",
	"IFLA_BRPORT_MESSAGE_AGE_TIMER",
	"IFLA_BRPORT_FORWARD_DELAY_TIMER",
	"IFLA_BRPORT_HOLD_TIMER",
	"IFLA_BRPORT_FLUSH",
	"IFLA_BRPORT_MULTICAST_ROUTER",
	"IFLA_BRPORT_PAD",
	"IFLA_BRPORT_MCAST_FLOOD",
	"IFLA_BRPORT_MCAST_TO_UCAST",
	"IFLA_BRPORT_VLAN_TUNNEL",
	"IFLA_BRPORT_BCAST_FLOOD",
	"IFLA_BRPORT_GROUP_FWD_MASK",
	"IFLA_BRPORT_NEIGH_SUPPRESS",
	"IFLA_BRPORT_ISOLATED",
	"IFLA_BRPORT_BACKUP_PORT",
	"IFLA_BRPORT_MRP_RING_OPEN",
	"IFLA_BRPORT_MRP_IN_OPEN",
	"IFLA_BRPORT_MCAST_EHT_HOSTS_LIMIT",
	"IFLA_BRPORT_MCAST_EHT_HOSTS_CNT",
	"IFLA_BRPORT_LOCKED",
	"IFLA_BRPORT_MAB",
	"IFLA_BRPORT_MCAST_N_GROUPS",
	"IFLA_BRPORT_MCAST_MAX_GROUPS",
	"IFLA_BRPORT_NEIGH_VLAN_SUPPRESS",
	"IFLA_BRPORT_BACKUP_NHID",
	"__IFLA_BRPORT_MAX"
};

static const char* IFLA_INFO[] = {
	"IFLA_INFO_UNSPEC",
	"IFLA_INFO_KIND",
	"IFLA_INFO_DATA",
	"IFLA_INFO_XSTATS",
	"IFLA_INFO_SLAVE_KIND",
	"IFLA_INFO_SLAVE_DATA",
	"__IFLA_INFO_MAX",
};

/* VLAN section */
static const char* IFLA_VLAN[] = {
	"IFLA_VLAN_UNSPEC",
	"IFLA_VLAN_ID",
	"IFLA_VLAN_FLAGS",
	"IFLA_VLAN_EGRESS_QOS",
	"IFLA_VLAN_INGRESS_QOS",
	"IFLA_VLAN_PROTOCOL",
	"__IFLA_VLAN_MAX",
};
static const char* IFLA_VLAN_QOS[] = {
	"IFLA_VLAN_QOS_UNSPEC",
	"IFLA_VLAN_QOS_MAPPING",
	"__IFLA_VLAN_QOS_MAX",
};

/* MACVLAN section */
static const char* IFLA_MACVLAN[] = {
	"IFLA_MACVLAN_UNSPEC",
	"IFLA_MACVLAN_MODE",
	"IFLA_MACVLAN_FLAGS",
	"IFLA_MACVLAN_MACADDR_MODE",
	"IFLA_MACVLAN_MACADDR",
	"IFLA_MACVLAN_MACADDR_DATA",
	"IFLA_MACVLAN_MACADDR_COUNT",
	"IFLA_MACVLAN_BC_QUEUE_LEN",
	"IFLA_MACVLAN_BC_QUEUE_LEN_USED",
	"IFLA_MACVLAN_BC_CUTOFF",
	"__IFLA_MACVLAN_MAX",
};
static const char* MACVLAN_MACADDR[] = {
	"MACVLAN_MACADDR_ADD",
	"MACVLAN_MACADDR_DEL",
	"MACVLAN_MACADDR_FLUSH",
	"MACVLAN_MACADDR_SET"
};

/* VRF section */
static const char* IFLA_VRF[] = {
	"IFLA_VRF_UNSPEC",
	"IFLA_VRF_TABLE",
	"__IFLA_VRF_MAX",
};
static const char* IFLA_VRF_PORT[] = {
	"IFLA_VRF_PORT_UNSPEC",
	"IFLA_VRF_PORT_TABLE",
	"__IFLA_VRF_PORT_MAX",
};

/* MACSEC section */
static const char* IFLA_MACSEC[] = {
	"IFLA_MACSEC_UNSPEC",
	"IFLA_MACSEC_SCI",
	"IFLA_MACSEC_PORT",
	"IFLA_MACSEC_ICV_LEN",
	"IFLA_MACSEC_CIPHER_SUITE",
	"IFLA_MACSEC_WINDOW",
	"IFLA_MACSEC_ENCODING_SA",
	"IFLA_MACSEC_ENCRYPT",
	"IFLA_MACSEC_PROTECT",
	"IFLA_MACSEC_INC_SCI",
	"IFLA_MACSEC_ES",
	"IFLA_MACSEC_SCB",
	"IFLA_MACSEC_REPLAY_PROTECT",
	"IFLA_MACSEC_VALIDATION",
	"IFLA_MACSEC_PAD",
	"IFLA_MACSEC_OFFLOAD",
	"__IFLA_MACSEC_MAX",
};

/* XFRM section */
static const char* IFLA_XFRM[] = {
	"IFLA_XFRM_UNSPEC",
	"IFLA_XFRM_LINK",
	"IFLA_XFRM_IF_ID",
	"IFLA_XFRM_COLLECT_METADATA",
	"__IFLA_XFRM_MAX"
};

/* IPVLAN section */
static const char* IFLA_IPVLAN[] = {
	"IFLA_IPVLAN_UNSPEC",
	"IFLA_IPVLAN_MODE",
	"IFLA_IPVLAN_FLAGS",
	"__IFLA_IPVLAN_MAX"
};
static const char* IPVLAN_MODE[] = {
	"IPVLAN_MODE_L2",
	"IPVLAN_MODE_L3",
	"IPVLAN_MODE_L3S",
	"IPVLAN_MODE_MAX"
};

static const char* NETKIT_MODE[] = {
	"NETKIT_L2",
	"NETKIT_L3",
};

static const char* IFLA_NETKIT[] = {
	"IFLA_NETKIT_UNSPEC",
	"IFLA_NETKIT_PEER_INFO",
	"IFLA_NETKIT_PRIMARY",
	"IFLA_NETKIT_POLICY",
	"IFLA_NETKIT_PEER_POLICY",
	"IFLA_NETKIT_MODE",
	"__IFLA_NETKIT_MAX",
};

/* VXLAN section */
static const char* VNIFILTER_ENTRY[] = {
	"VNIFILTER_ENTRY_STATS_UNSPEC",
	"VNIFILTER_ENTRY_STATS_RX_BYTES",
	"VNIFILTER_ENTRY_STATS_RX_PKTS",
	"VNIFILTER_ENTRY_STATS_RX_DROPS",
	"VNIFILTER_ENTRY_STATS_RX_ERRORS",
	"VNIFILTER_ENTRY_STATS_TX_BYTES",
	"VNIFILTER_ENTRY_STATS_TX_PKTS",
	"VNIFILTER_ENTRY_STATS_TX_DROPS",
	"VNIFILTER_ENTRY_STATS_TX_ERRORS",
	"VNIFILTER_ENTRY_STATS_PAD",
	"__VNIFILTER_ENTRY_STATS_MAX"
};

static const char* VXLAN_VNIFILTERENTRY[] = {
	"VXLAN_VNIFILTER_ENTRY_UNSPEC",
	"VXLAN_VNIFILTER_ENTRY_START",
	"VXLAN_VNIFILTER_ENTRY_END",
	"VXLAN_VNIFILTER_ENTRY_GROUP",
	"VXLAN_VNIFILTER_ENTRY_GROUP6",
	"VXLAN_VNIFILTER_ENTRY_STATS",
	"__VXLAN_VNIFILTER_ENTRY_MAX"
};

static const char* VXLAN_VNIFILTER[] = {
	"VXLAN_VNIFILTER_UNSPEC",
	"VXLAN_VNIFILTER_ENTRY",
	"__VXLAN_VNIFILTER_MAX"
};

static const char* IFLA_VXLAN[] = {
	"IFLA_VXLAN_UNSPEC",
	"IFLA_VXLAN_ID",
	"IFLA_VXLAN_GROUP",	/* group or remote address */
	"IFLA_VXLAN_LINK",
	"IFLA_VXLAN_LOCAL",
	"IFLA_VXLAN_TTL",
	"IFLA_VXLAN_TOS",
	"IFLA_VXLAN_LEARNING",
	"IFLA_VXLAN_AGEING",
	"IFLA_VXLAN_LIMIT",
	"IFLA_VXLAN_PORT_RANGE",	/* source port */
	"IFLA_VXLAN_PROXY",
	"IFLA_VXLAN_RSC",
	"IFLA_VXLAN_L2MISS",
	"IFLA_VXLAN_L3MISS",
	"IFLA_VXLAN_PORT",	/* destination port */
	"IFLA_VXLAN_GROUP6",
	"IFLA_VXLAN_LOCAL6",
	"IFLA_VXLAN_UDP_CSUM",
	"IFLA_VXLAN_UDP_ZERO_CSUM6_TX",
	"IFLA_VXLAN_UDP_ZERO_CSUM6_RX",
	"IFLA_VXLAN_REMCSUM_TX",
	"IFLA_VXLAN_REMCSUM_RX",
	"IFLA_VXLAN_GBP",
	"IFLA_VXLAN_REMCSUM_NOPARTIAL",
	"IFLA_VXLAN_COLLECT_METADATA",
	"IFLA_VXLAN_LABEL",
	"IFLA_VXLAN_GPE",
	"IFLA_VXLAN_TTL_INHERIT",
	"IFLA_VXLAN_DF",
	"IFLA_VXLAN_VNIFILTER", /* only applicable with COLLECT_METADATA mode */
	"IFLA_VXLAN_LOCALBYPASS",
	"IFLA_VXLAN_LABEL_POLICY", /* IPv6 flow label policy; ifla_vxlan_label_policy */
	"__IFLA_VXLAN_MAX"
};

static const char* IFLA_VXLANDF[] =  {
	"VXLAN_DF_UNSET",
	"VXLAN_DF_SET",
	"VXLAN_DF_INHERIT",
	"__VXLAN_DF_END",
};

/*
static const char* IFLA_VXLAN_LABEL_POLICY[] = {
	VXLAN_LABEL_FIXED ,
	VXLAN_LABEL_INHERIT,
	__VXLAN_LABEL_END,
	VXLAN_LABEL_MAX = __VXLAN_LABEL_END - 1,
};
*/

/* GENEVE section */
static const char* IFLA_GENEVE[] = {
	"IFLA_GENEVE_UNSPEC",
	"IFLA_GENEVE_ID",
	"IFLA_GENEVE_REMOTE",
	"IFLA_GENEVE_TTL",
	"IFLA_GENEVE_TOS",
	"IFLA_GENEVE_PORT",	/* destination port */
	"IFLA_GENEVE_COLLECT_METADATA",
	"IFLA_GENEVE_REMOTE6",
	"IFLA_GENEVE_UDP_CSUM",
	"IFLA_GENEVE_UDP_ZERO_CSUM6_TX",
	"IFLA_GENEVE_UDP_ZERO_CSUM6_RX",
	"IFLA_GENEVE_LABEL",
	"IFLA_GENEVE_TTL_INHERIT",
	"IFLA_GENEVE_DF",
	"IFLA_GENEVE_INNER_PROTO_INHERIT",
	"__IFLA_GENEVE_MAX"
};
static const char* IFLA_GENEVEDF[] = {
	"GENEVE_DF_UNSET",
	"GENEVE_DF_SET",
	"GENEVE_DF_INHERIT",
	"__GENEVE_DF_END"
};

/* Bareudp section  */
static const char* IFLA_BAREUDP[] = {
	"IFLA_BAREUDP_UNSPEC",
	"IFLA_BAREUDP_PORT",
	"IFLA_BAREUDP_ETHERTYPE",
	"IFLA_BAREUDP_SRCPORT_MIN",
	"IFLA_BAREUDP_MULTIPROTO_MODE",
	"__IFLA_BAREUDP_MAX"
};

/* PPP section */
static const char* IFLA_PPP[] = {
	"IFLA_PPP_UNSPEC",
	"IFLA_PPP_DEV_FD",
	"__IFLA_PPP_MAX"
};

/* GTP section */

static const char* GTP_ROLE[] = {
	"GTP_ROLE_GGSN",
	"GTP_ROLE_SGSN",
};

static const char* IFLA_GTP[] = {
	"IFLA_GTP_UNSPEC",
	"IFLA_GTP_FD0",
	"IFLA_GTP_FD1",
	"IFLA_GTP_PDP_HASHSIZE",
	"IFLA_GTP_ROLE",
	"IFLA_GTP_CREATE_SOCKETS",
	"IFLA_GTP_RESTART_COUNT",
	"__IFLA_GTP_MAX",
};


/* Bonding section */

static const char* IFLA_BOND[] = {
	"IFLA_BOND_UNSPEC",
	"IFLA_BOND_MODE",
	"IFLA_BOND_ACTIVE_SLAVE",
	"IFLA_BOND_MIIMON",
	"IFLA_BOND_UPDELAY",
	"IFLA_BOND_DOWNDELAY",
	"IFLA_BOND_USE_CARRIER",
	"IFLA_BOND_ARP_INTERVAL",
	"IFLA_BOND_ARP_IP_TARGET",
	"IFLA_BOND_ARP_VALIDATE",
	"IFLA_BOND_ARP_ALL_TARGETS",
	"IFLA_BOND_PRIMARY",
	"IFLA_BOND_PRIMARY_RESELECT",
	"IFLA_BOND_FAIL_OVER_MAC",
	"IFLA_BOND_XMIT_HASH_POLICY",
	"IFLA_BOND_RESEND_IGMP",
	"IFLA_BOND_NUM_PEER_NOTIF",
	"IFLA_BOND_ALL_SLAVES_ACTIVE",
	"IFLA_BOND_MIN_LINKS",
	"IFLA_BOND_LP_INTERVAL",
	"IFLA_BOND_PACKETS_PER_SLAVE",
	"IFLA_BOND_AD_LACP_RATE",
	"IFLA_BOND_AD_SELECT",
	"IFLA_BOND_AD_INFO",
	"IFLA_BOND_AD_ACTOR_SYS_PRIO",
	"IFLA_BOND_AD_USER_PORT_KEY",
	"IFLA_BOND_AD_ACTOR_SYSTEM",
	"IFLA_BOND_TLB_DYNAMIC_LB",
	"IFLA_BOND_PEER_NOTIF_DELAY",
	"IFLA_BOND_AD_LACP_ACTIVE",
	"IFLA_BOND_MISSED_MAX",
	"IFLA_BOND_NS_IP6_TARGET",
	"IFLA_BOND_COUPLED_CONTROL",
	"__IFLA_BOND_MAX",
};
static const char* IFLA_BOND_ADINFO[] = {
	"IFLA_BOND_AD_INFO_UNSPEC",
	"IFLA_BOND_AD_INFO_AGGREGATOR",
	"IFLA_BOND_AD_INFO_NUM_PORTS",
	"IFLA_BOND_AD_INFO_ACTOR_KEY",
	"IFLA_BOND_AD_INFO_PARTNER_KEY",
	"IFLA_BOND_AD_INFO_PARTNER_MAC",
	"__IFLA_BOND_AD_INFO_MAX",
};

static const char* IFLA_BOND_SLAVE[] = {
	"IFLA_BOND_SLAVE_UNSPEC",
	"IFLA_BOND_SLAVE_STATE",
	"IFLA_BOND_SLAVE_MII_STATUS",
	"IFLA_BOND_SLAVE_LINK_FAILURE_COUNT",
	"IFLA_BOND_SLAVE_PERM_HWADDR",
	"IFLA_BOND_SLAVE_QUEUE_ID",
	"IFLA_BOND_SLAVE_AD_AGGREGATOR_ID",
	"IFLA_BOND_SLAVE_AD_ACTOR_OPER_PORT_STATE",
	"IFLA_BOND_SLAVE_AD_PARTNER_OPER_PORT_STATE",
	"IFLA_BOND_SLAVE_PRIO",
	"__IFLA_BOND_SLAVE_MAX",
};

/* SR-IOV virtual function management section */

static const char* IFLA_VFINFO[] = {
	"IFLA_VF_INFO_UNSPEC",
	"IFLA_VF_INFO",
	"__IFLA_VF_INFO_MAX",
};

static const char* IFLA_VF[] = {
	"IFLA_VF_UNSPEC",
	"IFLA_VF_MAC",		/* Hardware queue specific attributes */
	"IFLA_VF_VLAN",		/* VLAN ID and QoS */
	"IFLA_VF_TX_RATE",	/* Max TX Bandwidth Allocation */
	"IFLA_VF_SPOOFCHK",	/* Spoof Checking on/off switch */
	"IFLA_VF_LINK_STATE",	/* link state enable/disable/auto switch */
	"IFLA_VF_RATE",		/* Min and Max TX Bandwidth Allocation */
	"IFLA_VF_RSS_QUERY_EN",	/* RSS Redirection Table and Hash Key query
				 * on/off switch
				 */
	"IFLA_VF_STATS",		/* network device statistics */
	"IFLA_VF_TRUST",		/* Trust VF */
	"IFLA_VF_IB_NODE_GUID",	/* VF Infiniband node GUID */
	"IFLA_VF_IB_PORT_GUID",	/* VF Infiniband port GUID */
	"IFLA_VF_VLAN_LIST",	/* nested list of vlans, option for QinQ */
	"IFLA_VF_BROADCAST",	/* VF broadcast */
	"__IFLA_VF_MAX",
};

static const char* IFLA_VF_VLANINFO[] = {
	"IFLA_VF_VLAN_INFO_UNSPEC",
	"IFLA_VF_VLAN_INFO",	/* VLAN ID, QoS and VLAN protocol */
	"__IFLA_VF_VLAN_INFO_MAX",
};

static const char* IFLA_VF_LINKSTATE[] = {
	"IFLA_VF_LINK_STATE_AUTO",	/* link state of the uplink */
	"IFLA_VF_LINK_STATE_ENABLE",	/* link always up */
	"IFLA_VF_LINK_STATE_DISABLE",	/* link always down */
	"__IFLA_VF_LINK_STATE_MAX",
};

static const char* IFLA_VFSTATS[] = {
	"IFLA_VF_STATS_RX_PACKETS",
	"IFLA_VF_STATS_TX_PACKETS",
	"IFLA_VF_STATS_RX_BYTES",
	"IFLA_VF_STATS_TX_BYTES",
	"IFLA_VF_STATS_BROADCAST",
	"IFLA_VF_STATS_MULTICAST",
	"IFLA_VF_STATS_PAD",
	"IFLA_VF_STATS_RX_DROPPED",
	"IFLA_VF_STATS_TX_DROPPED",
	"__IFLA_VF_STATS_MAX",
};
/* VF ports management section */

static const char* IFLA_VFPORT[] = {
	"IFLA_VF_PORT_UNSPEC",
	"IFLA_VF_PORT",			/* nest */
	"__IFLA_VF_PORT_MAX",
};

static const char* IFLA_PORT[] = {
	"IFLA_PORT_UNSPEC",
	"IFLA_PORT_VF",			/* __u32 */
	"IFLA_PORT_PROFILE",		/* string */
	"IFLA_PORT_VSI_TYPE",		/* 802.1Qbg (pre-)standard VDP */
	"IFLA_PORT_INSTANCE_UUID",	/* binary UUID */
	"IFLA_PORT_HOST_UUID",		/* binary UUID */
	"IFLA_PORT_REQUEST",		/* __u8 */
	"IFLA_PORT_RESPONSE",		/* __u16, output only */
	"__IFLA_PORT_MAX",
};

static const char* PORT_REQUEST[] = {
	"PORT_REQUEST_PREASSOCIATE",
	"PORT_REQUEST_PREASSOCIATE_RR",
	"PORT_REQUEST_ASSOCIATE",
	"PORT_REQUEST_DISASSOCIATE",
};

static const char* PORT_VDP_RESPONSE[] = {
	"PORT_VDP_RESPONSE_SUCCESS",
	"PORT_VDP_RESPONSE_INVALID_FORMAT",
	"PORT_VDP_RESPONSE_INSUFFICIENT_RESOURCES",
	"PORT_VDP_RESPONSE_UNUSED_VTID",
	"PORT_VDP_RESPONSE_VTID_VIOLATION",
	"PORT_VDP_RESPONSE_VTID_VERSION_VIOALTION",
	"PORT_VDP_RESPONSE_OUT_OF_SYNC",
};

/* IPoIB section */

static const char* IFLA_IPOIB[] = {
	"IFLA_IPOIB_UNSPEC",
	"IFLA_IPOIB_PKEY",
	"IFLA_IPOIB_MODE",
	"IFLA_IPOIB_UMCAST",
	"__IFLA_IPOIB_MAX"
};

static const char* IPOIB_MODE[] = {
	"IPOIB_MODE_DATAGRAM", /* using unreliable datagram QPs */
	"IPOIB_MODE_CONNECTED" /* using connected QPs */
};

/* HSR/PRP section, both uses same interface */

/* Different redundancy protocols for hsr device */
static const char* HSR_PROTOCOL[] = {
	"HSR_PROTOCOL_HSR",
	"HSR_PROTOCOL_PRP",
	"HSR_PROTOCOL_MAX",
};

static const char* IFLA_HSR[] = {
	"IFLA_HSR_UNSPEC",
	"IFLA_HSR_SLAVE1",
	"IFLA_HSR_SLAVE2",
	"IFLA_HSR_MULTICAST_SPEC",	/* Last byte of supervision addr */
	"IFLA_HSR_SUPERVISION_ADDR",	/* Supervision frame multicast addr */
	"IFLA_HSR_SEQ_NR",
	"IFLA_HSR_VERSION",		/* HSR version */
	"IFLA_HSR_PROTOCOL",		/* Indicate different protocol than
					 * HSR. For example PRP.
					 */
	"__IFLA_HSR_MAX",
};

static const char* IFLASTATS[] = {
	"IFLA_STATS_UNSPEC", /* also used as 64bit pad attribute */
	"IFLA_STATS_LINK_64",
	"IFLA_STATS_LINK_XSTATS",
	"IFLA_STATS_LINK_XSTATS_SLAVE",
	"IFLA_STATS_LINK_OFFLOAD_XSTATS",
	"IFLA_STATS_AF_SPEC",
	"__IFLA_STATS_MAX",
};

static const char* IFLA_STATS_GET_SET[] = {
	"IFLA_STATS_GETSET_UNSPEC",
	"IFLA_STATS_GET_FILTERS", /* Nest of IFLA_STATS_LINK_xxx, each a u32 with
				 * a filter mask for the corresponding group.
				 */
	"IFLA_STATS_SET_OFFLOAD_XSTATS_L3_STATS", /* 0 or 1 as u8 */
	"__IFLA_STATS_GETSET_MAX",
};

/* These are embedded into IFLA_STATS_LINK_XSTATS:
 * [IFLA_STATS_LINK_XSTATS]
 * -> [LINK_XSTATS_TYPE_xxx]
 *    -> [rtnl link type specific attributes]
 */
static const char* LINK_XSTATS_TYPE[] = {
	"LINK_XSTATS_TYPE_UNSPEC",
	"LINK_XSTATS_TYPE_BRIDGE",
	"LINK_XSTATS_TYPE_BOND",
	"__LINK_XSTATS_TYPE_MAX"
};

/* These are stats embedded into IFLA_STATS_LINK_OFFLOAD_XSTATS */
static const char* IFLA_OFFLOAD_XSTATS[] = {
	"IFLA_OFFLOAD_XSTATS_UNSPEC",
	"IFLA_OFFLOAD_XSTATS_CPU_HIT", /* struct rtnl_link_stats64 */
	"IFLA_OFFLOAD_XSTATS_HW_S_INFO",	/* HW stats info. A nest */
	"IFLA_OFFLOAD_XSTATS_L3_STATS",	/* struct rtnl_hw_stats64 */
	"__IFLA_OFFLOAD_XSTATS_MAX"
};

static const char* IFLA_OFFLOAD_XSTATS_HW_SINFO[] = {
	"IFLA_OFFLOAD_XSTATS_HW_S_INFO_UNSPEC",
	"IFLA_OFFLOAD_XSTATS_HW_S_INFO_REQUEST",		/* u8 */
	"IFLA_OFFLOAD_XSTATS_HW_S_INFO_USED",		/* u8 */
	"__IFLA_OFFLOAD_XSTATS_HW_S_INFO_MAX",
};

/* These are stored into IFLA_XDP_ATTACHED on dump. */
static const char* XDP_ATTACHED[] = {
	"XDP_ATTACHED_NONE",
	"XDP_ATTACHED_DRV",
	"XDP_ATTACHED_SKB",
	"XDP_ATTACHED_HW",
	"XDP_ATTACHED_MULTI",
};

static const char* IFLAXDP[] = {
	"IFLA_XDP_UNSPEC",
	"IFLA_XDP_FD",
	"IFLA_XDP_ATTACHED",
	"IFLA_XDP_FLAGS",
	"IFLA_XDP_PROG_ID",
	"IFLA_XDP_DRV_PROG_ID",
	"IFLA_XDP_SKB_PROG_ID",
	"IFLA_XDP_HW_PROG_ID",
	"IFLA_XDP_EXPECTED_FD",
	"__IFLA_XDP_MAX",
};


static const char* IFLAEVENT[] = {
	"IFLA_EVENT_NONE",
	"IFLA_EVENT_REBOOT",		/* internal reset / reboot */
	"IFLA_EVENT_FEATURES",		/* change in offload features */
	"IFLA_EVENT_BONDING_FAILOVER",	/* change in active slave */
	"IFLA_EVENT_NOTIFY_PEERS",	/* re-sent grat. arp/ndisc */
	"IFLA_EVENT_IGMP_RESEND",		/* re-sent IGMP JOIN */
	"IFLA_EVENT_BONDING_OPTIONS",	/* change in bonding options */
};

/* tun section */

static const char* IFLA_TUN[] = {
	"IFLA_TUN_UNSPEC",
	"IFLA_TUN_OWNER",
	"IFLA_TUN_GROUP",
	"IFLA_TUN_TYPE",
	"IFLA_TUN_PI",
	"IFLA_TUN_VNET_HDR",
	"IFLA_TUN_PERSIST",
	"IFLA_TUN_MULTI_QUEUE",
	"IFLA_TUN_NUM_QUEUES",
	"IFLA_TUN_NUM_DISABLED_QUEUES",
	"__IFLA_TUN_MAX",
};

/* rmnet section */

static const char* IFLA_RMNET[] = {
	"IFLA_RMNET_UNSPEC",
	"IFLA_RMNET_MUX_ID",
	"IFLA_RMNET_FLAGS",
	"__IFLA_RMNET_MAX",
};

/* MCTP section */

static const char* IFLA_MCTP[] = {
	"IFLA_MCTP_UNSPEC",
	"IFLA_MCTP_NET",
	"__IFLA_MCTP_MAX",
};

 /* DSA section */

static const char* IFLA_DSA[] = {
	"IFLA_DSA_UNSPEC",
	"IFLA_DSA_CONDUIT",
	"__IFLA_DSA_MAX",
};


static PyMethodDef link_add_methods[] = {
    {NULL, NULL, 0, NULL}
};

static struct PyModuleDef link_add_module = {
    PyModuleDef_HEAD_INIT,
    "_const_linkadd",   /* name of module */
    const_linkadd_doc,  /* module documentation, may be NULL */
    -1,            /* size of per-interpreter state of the module,
                      or -1 if the module keeps state in global variables. */
    link_add_methods
};


PyMODINIT_FUNC PyInit__const_linkadd(void){
    PyObject* module = PyModule_Create(&link_add_module);

	ADD_ENUM(IFLA, __IFLA_MAX);
	ADD_INT("IFLA_MAX", IFLA_MAX);
	ADD_INT("IFLA_TARGET_NETNSID", IFLA_TARGET_NETNSID);

	ADD_ENUM(IFLA_PROTO_DOWNREASON, __IFLA_PROTO_DOWN_REASON_CNT);
	ADD_INT("IFLA_PROTO_DOWN_REASON_MAX", IFLA_PROTO_DOWN_REASON_MAX);

	ADD_ENUM(IFLA_INET, __IFLA_INET_MAX);
	ADD_INT("IFLA_INET_MAX", IFLA_INET_MAX);
	ADD_ENUM(IFLA_INET6, __IFLA_INET6_MAX);
	ADD_INT("IFLA_INET6_MAX", IFLA_INET6_MAX);
	ADD_ENUM(IFLA_INET6_GEN_MODE,IN6_ADDR_GEN_MODE_RANDOM);

	ADD_ENUM(IFLA_BR, __IFLA_BR_MAX);
	ADD_INT("IFLA_BR_MAX", IFLA_BR_MAX);

	ADD_ENUM(BRIDGE_MODE, BRIDGE_MODE_HAIRPIN);

	ADD_ENUM(IFLA_BRPORT, __IFLA_BRPORT_MAX);
	ADD_INT("IFLA_BRPORT_MAX", IFLA_BRPORT_MAX);

	ADD_ENUM(IFLA_INFO, __IFLA_INFO_MAX);
	ADD_INT("IFLA_INFO_MAX", IFLA_INFO_MAX);

	ADD_ENUM(IFLA_VLAN, __IFLA_VLAN_MAX);
	ADD_INT("IFLA_VLAN_MAX", IFLA_VLAN_MAX);

	ADD_ENUM(IFLA_VLAN_QOS, __IFLA_VLAN_QOS_MAX);
	ADD_INT("IFLA_VLAN_QOS_MAX",IFLA_VLAN_QOS_MAX);

	ADD_ENUM(IFLA_MACVLAN,__IFLA_MACVLAN_MAX);
	ADD_INT("IFLA_MACVLAN_MAX", IFLA_MACVLAN_MAX);

	ADD_INT("MACVLAN_MODE_PRIVATE", MACVLAN_MODE_PRIVATE);
	ADD_INT("MACVLAN_MODE_VEPA", MACVLAN_MODE_VEPA);
	ADD_INT("MACVLAN_MODE_BRIDGE", MACVLAN_MODE_BRIDGE);
	ADD_INT("MACVLAN_MODE_PASSTHRU", MACVLAN_MODE_PASSTHRU);
	ADD_INT("MACVLAN_MODE_SOURCE", MACVLAN_MODE_SOURCE);

	ADD_ENUM(MACVLAN_MACADDR, MACVLAN_MACADDR_SET);
	ADD_INT("MACVLAN_FLAG_NOPROMISC", MACVLAN_FLAG_NOPROMISC);
	ADD_INT("MACVLAN_FLAG_NODST", MACVLAN_FLAG_NODST);

	ADD_ENUM(IFLA_VRF, __IFLA_VRF_MAX);
	ADD_INT("IFLA_VRF_MAX", IFLA_VRF_MAX);

	ADD_ENUM(IFLA_VRF_PORT, __IFLA_VRF_PORT_MAX);
	ADD_INT("IFLA_VRF_PORT_MAX", IFLA_VRF_PORT_MAX);

	ADD_ENUM(IFLA_MACSEC, __IFLA_MACSEC_MAX);
	ADD_INT("IFLA_MACSEC_MAX", IFLA_MACSEC_MAX);

	ADD_ENUM(IFLA_XFRM,__IFLA_XFRM_MAX);
	ADD_INT("IFLA_XFRM_MAX", IFLA_XFRM_MAX);
	ADD_INT("MACSEC_VALIDATE_DISABLED", MACSEC_VALIDATE_DISABLED);
	ADD_INT("MACSEC_VALIDATE_CHECK", MACSEC_VALIDATE_CHECK);
	ADD_INT("MACSEC_VALIDATE_STRICT", MACSEC_VALIDATE_STRICT);
	ADD_INT("__MACSEC_VALIDATE_END", __MACSEC_VALIDATE_END);
	ADD_INT("MACSEC_VALIDATE_MAX", MACSEC_VALIDATE_MAX);
	ADD_INT("MACSEC_OFFLOAD_OFF",MACSEC_OFFLOAD_OFF);
	ADD_INT("MACSEC_OFFLOAD_PHY",MACSEC_OFFLOAD_PHY);
	ADD_INT("MACSEC_OFFLOAD_MAC",MACSEC_OFFLOAD_MAC);
	ADD_INT("__MACSEC_OFFLOAD_END",__MACSEC_OFFLOAD_END);
	ADD_INT("MACSEC_OFFLOAD_MAX",MACSEC_OFFLOAD_MAX);

	ADD_ENUM(IFLA_IPVLAN,__IFLA_IPVLAN_MAX);
	ADD_INT("IFLA_IPVLAN_MAX", IFLA_IPVLAN_MAX);

	ADD_ENUM(IPVLAN_MODE,IPVLAN_MODE_MAX);
	ADD_INT("IPVLAN_F_PRIVATE",IPVLAN_F_PRIVATE);
	ADD_INT("IPVLAN_F_VEPA", IPVLAN_F_VEPA);

	ADD_INT("NETKIT_NEXT",NETKIT_NEXT);
	ADD_INT("NETKIT_PASS",NETKIT_PASS);
	ADD_INT("NETKIT_DROP",NETKIT_DROP);
	ADD_INT("NETKIT_REDIRECT",NETKIT_REDIRECT);
	ADD_ENUM(NETKIT_MODE, NETKIT_L3);
	ADD_ENUM(IFLA_NETKIT, __IFLA_NETKIT_MAX);
	ADD_INT("IFLA_NETKIT_MAX", IFLA_NETKIT_MAX);

	ADD_INT("TUNNEL_MSG_FLAG_STATS", TUNNEL_MSG_FLAG_STATS);
	ADD_INT("TUNNEL_MSG_VALID_USER_FLAGS", TUNNEL_MSG_VALID_USER_FLAGS);
	ADD_ENUM(VNIFILTER_ENTRY, __VNIFILTER_ENTRY_STATS_MAX);
	ADD_INT("VNIFILTER_ENTRY_STATS_MAX", VNIFILTER_ENTRY_STATS_MAX);

	ADD_ENUM(VXLAN_VNIFILTERENTRY, __VXLAN_VNIFILTER_ENTRY_MAX);
	ADD_INT("VXLAN_VNIFILTER_ENTRY_MAX", VXLAN_VNIFILTER_ENTRY_MAX);
	ADD_ENUM(VXLAN_VNIFILTER, __VXLAN_VNIFILTER_MAX);
	ADD_INT("VXLAN_VNIFILTER_MAX", VXLAN_VNIFILTER_MAX);

	ADD_ENUM(IFLA_VXLAN, __IFLA_VXLAN_MAX);
	ADD_INT("IFLA_VXLAN_MAX", IFLA_VXLAN_MAX);
	ADD_ENUM(IFLA_VXLANDF, __VXLAN_DF_END);
	ADD_INT("VXLAN_DF_MAX", VXLAN_DF_MAX);

	ADD_ENUM(IFLA_GENEVE, __IFLA_GENEVE_MAX);
	ADD_INT("IFLA_GENEVE_MAX",IFLA_GENEVE_MAX);
	ADD_ENUM(IFLA_GENEVEDF, __GENEVE_DF_END);
	ADD_INT("GENEVE_DF_MAX", GENEVE_DF_MAX);

	ADD_ENUM(IFLA_BAREUDP, __IFLA_BAREUDP_MAX);
	ADD_INT("IFLA_BAREUDP_MAX", IFLA_BAREUDP_MAX);

	ADD_ENUM(IFLA_PPP, __IFLA_PPP_MAX);
	ADD_INT("IFLA_PPP_MAX", IFLA_PPP_MAX);
	
	ADD_ENUM(GTP_ROLE, GTP_ROLE_SGSN + 1);
	ADD_ENUM(IFLA_GTP, __IFLA_GTP_MAX);
	ADD_INT("IFLA_GTP_MAX", IFLA_GTP_MAX);

	ADD_ENUM(IFLA_BOND, __IFLA_BOND_MAX);
	ADD_INT("IFLA_BOND_MAX", IFLA_BOND_MAX);
	ADD_ENUM(IFLA_BOND_ADINFO, __IFLA_BOND_AD_INFO_MAX);
	ADD_INT("IFLA_BOND_AD_INFO_MAX", IFLA_BOND_AD_INFO_MAX);
	ADD_ENUM(IFLA_BOND_SLAVE, __IFLA_BOND_SLAVE_MAX);
	ADD_INT("IFLA_BOND_SLAVE_MAX", IFLA_BOND_SLAVE_MAX);

	ADD_ENUM(IFLA_VFINFO, __IFLA_VF_INFO_MAX);
	ADD_INT("IFLA_VF_INFO_MAX", IFLA_VF_INFO_MAX);
	ADD_ENUM(IFLA_VF, __IFLA_VF_MAX);
	ADD_INT("IFLA_VF_MAX", IFLA_VF_MAX);
	ADD_ENUM(IFLA_VF_VLANINFO, __IFLA_VF_VLAN_INFO_MAX);
	ADD_INT("IFLA_VF_VLAN_INFO_MAX", IFLA_VF_VLAN_INFO_MAX);
	ADD_INT("MAX_VLAN_LIST_LEN", MAX_VLAN_LIST_LEN);
	ADD_ENUM(IFLA_VF_LINKSTATE, __IFLA_VF_LINK_STATE_MAX);
	ADD_ENUM(IFLA_VFSTATS, __IFLA_VF_STATS_MAX);
	ADD_INT("IFLA_VF_STATS_MAX", IFLA_VF_STATS_MAX);
	ADD_ENUM(IFLA_VFPORT, __IFLA_VF_PORT_MAX);
	ADD_INT("IFLA_VF_PORT_MAX", IFLA_VF_PORT_MAX);

	ADD_ENUM(IFLA_PORT, __IFLA_PORT_MAX);
	ADD_INT("IFLA_PORT_MAX", IFLA_PORT_MAX);
	ADD_INT("PORT_PROFILE_MAX", PORT_PROFILE_MAX);
	ADD_INT("PORT_UUID_MAX", PORT_UUID_MAX);
	ADD_INT("PORT_SELF_VF", PORT_SELF_VF);
	ADD_ENUM(PORT_REQUEST, PORT_REQUEST_DISASSOCIATE);

	ADD_ENUM(PORT_VDP_RESPONSE,PORT_VDP_RESPONSE_OUT_OF_SYNC);
		/* 0x08-0xFF reserved for future VDP use */
	ADD_INT("PORT_PROFILE_RESPONSE_SUCCESS",PORT_PROFILE_RESPONSE_SUCCESS);
	ADD_INT("PORT_PROFILE_RESPONSE_INPROGRESS",PORT_PROFILE_RESPONSE_INPROGRESS);
	ADD_INT("PORT_PROFILE_RESPONSE_INVALID",PORT_PROFILE_RESPONSE_INVALID);
	ADD_INT("PORT_PROFILE_RESPONSE_BADSTATE",PORT_PROFILE_RESPONSE_BADSTATE);
	ADD_INT("PORT_PROFILE_RESPONSE_INSUFFICIENT_RESOURCES",PORT_PROFILE_RESPONSE_INSUFFICIENT_RESOURCES);
	ADD_INT("PORT_PROFILE_RESPONSE_ERROR",PORT_PROFILE_RESPONSE_ERROR);

	ADD_ENUM(IFLA_IPOIB,__IFLA_IPOIB_MAX);
	ADD_INT("IFLA_IPOIB_MAX", IFLA_IPOIB_MAX);
	ADD_ENUM(IPOIB_MODE,IPOIB_MODE_CONNECTED);

	ADD_ENUM(HSR_PROTOCOL,HSR_PROTOCOL_MAX);
	ADD_ENUM(IFLA_HSR, __IFLA_HSR_MAX);
	ADD_INT("IFLA_HSR_MAX", IFLA_HSR_MAX);

	ADD_ENUM(IFLASTATS, __IFLA_STATS_MAX);
	ADD_INT("IFLA_STATS_MAX", IFLA_STATS_MAX);
	ADD_ENUM(IFLA_STATS_GET_SET, __IFLA_STATS_GETSET_MAX);
	ADD_INT("IFLA_STATS_GETSET_MAX", IFLA_STATS_GETSET_MAX);

	ADD_ENUM(LINK_XSTATS_TYPE, __LINK_XSTATS_TYPE_MAX);
	ADD_INT("LINK_XSTATS_TYPE_MAX", LINK_XSTATS_TYPE_MAX);

	ADD_ENUM(IFLA_OFFLOAD_XSTATS, __IFLA_OFFLOAD_XSTATS_MAX);
	ADD_INT("IFLA_OFFLOAD_XSTATS_MAX", IFLA_OFFLOAD_XSTATS_MAX);

	ADD_ENUM(IFLA_OFFLOAD_XSTATS_HW_SINFO, __IFLA_OFFLOAD_XSTATS_HW_S_INFO_MAX);
	ADD_INT("IFLA_OFFLOAD_XSTATS_HW_S_INFO_MAX", IFLA_OFFLOAD_XSTATS_HW_S_INFO_MAX);

	ADD_INT("XDP_FLAGS_UPDATE_IF_NOEXIST", XDP_FLAGS_UPDATE_IF_NOEXIST);
	ADD_INT("XDP_FLAGS_SKB_MODE",XDP_FLAGS_SKB_MODE);
	ADD_INT("XDP_FLAGS_DRV_MODE",XDP_FLAGS_DRV_MODE);
	ADD_INT("XDP_FLAGS_HW_MODE",XDP_FLAGS_HW_MODE);
	ADD_INT("XDP_FLAGS_REPLACE",XDP_FLAGS_REPLACE);
	ADD_INT("XDP_FLAGS_MODES",XDP_FLAGS_MODES);
	ADD_INT("XDP_FLAGS_MASK",XDP_FLAGS_MASK);
	ADD_ENUM(XDP_ATTACHED, XDP_ATTACHED_MULTI);
	ADD_ENUM(IFLAXDP, __IFLA_XDP_MAX);
	ADD_INT("IFLA_XDP_MAX", IFLA_XDP_MAX);
	ADD_ENUM(IFLAEVENT, IFLA_EVENT_BONDING_OPTIONS);

	ADD_ENUM(IFLA_TUN, __IFLA_TUN_MAX);
	ADD_INT("IFLA_TUN_MAX",IFLA_TUN_MAX);

	ADD_INT("RMNET_FLAGS_INGRESS_DEAGGREGATION",RMNET_FLAGS_INGRESS_DEAGGREGATION);
	ADD_INT("RMNET_FLAGS_INGRESS_MAP_COMMANDS",RMNET_FLAGS_INGRESS_MAP_COMMANDS);
	ADD_INT("RMNET_FLAGS_INGRESS_MAP_CKSUMV4",RMNET_FLAGS_INGRESS_MAP_CKSUMV4);
	ADD_INT("RMNET_FLAGS_EGRESS_MAP_CKSUMV4",RMNET_FLAGS_EGRESS_MAP_CKSUMV4);
	ADD_INT("RMNET_FLAGS_INGRESS_MAP_CKSUMV5",RMNET_FLAGS_INGRESS_MAP_CKSUMV5);
	ADD_INT("RMNET_FLAGS_EGRESS_MAP_CKSUMV5",RMNET_FLAGS_EGRESS_MAP_CKSUMV5);
	ADD_ENUM(IFLA_RMNET, __IFLA_RMNET_MAX);
	ADD_INT("IFLA_RMNET_MAX", IFLA_RMNET_MAX);

	ADD_ENUM(IFLA_MCTP, __IFLA_MCTP_MAX);
	ADD_INT("IFLA_MCTP_MAX", IFLA_MCTP_MAX);

	ADD_ENUM(IFLA_DSA, __IFLA_DSA_MAX);
	ADD_INT("IFLA_DSA_MAX", IFLA_DSA_MAX);

	ADD_INT("IFLA_VDE_VNL", IFLA_VDE_VNL);

	return module;
}